/* Lisp mode for Epsilon.  This editor mode is intended to be used with
 programs written in Common Lisp or Scheme.  It attempts to be compatible
 with Symbolics Zmacs and GNU Emacs.

 (c) Copyright 1990 Carl W. Hoffman.  All rights reserved.

 This file may be freely copied, distributed, or modified for non-commercial
 use provided that this copyright notice is not removed.  For further
 information about other Common Lisp and Scheme utilities, contact the
 following address:

 Carl W. Hoffman, 363 Marlborough Street, Boston, MA 02115, U.S.A.
 Internet: CWH@AI.MIT.EDU    CompuServe: 76416,3365    Fax: 617-262-4284

 This code has been tested with Epsilon version 6.0

 This file expects that FILLPREF.E will also be loaded so as to enable the
 filling of Lisp comments.

 shane@ai.mit.edu:  I gutted the dependency on fillpref.e and added the 
 ability to tag lisp functions.

 09/24/92 02:52: shane@ai.mit.edu:  Fix indenting bug caused by Lugaru
 changing to_column.  Incorporate fix for tagging bug from 
 Bill Cohagan <74375.313@compuserve.com>.

 30 December 2010: Lugaru added auto-highlighting parens, syntax
 highlighting, comment filling.
 */

#include <eel.h>
#include <colcode.h>

#define OPEN_PAREN  '('
#define CLOSE_PAREN ')'

/* buffer char fill_prefix[60]; */
keytable lisp_tab;

#define NOT_WHITESPACE           "[^ \t\n]"
#define NOT_WHITESPACE_OR_OPEN   "[^ \t\n(]"
#define NOT_WHITESPACE_OR_CLOSE  "[^ \t\n)]"
#define LISP_BREAK               "[^\\][ \t\n\"|()]"

char lisp_tag_pattern[64] = "^%(def|%(de</word>";
char lisp_word_pattern[64] = "[-A-Za-z0-9%$?:]+";

user char auto_show_lisp_delimiters = 1;
user char lisp_auto_show_delim_chars[20] = "{[]}";

color_class lisp_punctuation = color_class c_punctuation;
color_class lisp_comment = color_class c_comment;
color_class lisp_string = color_class c_string;
color_class lisp_function = color_class c_function;
color_class lisp_special = color_class c_preprocessor;

user buffer int lisp_indent = 2;

my_to_column(place)
{
  int start;
  
  to_begin_line ();
  start = point;
  {
    int end = point + parse_string (1, "[ \t]+");
    
    if (end > start)
      delete (start, end);
  }
  point = start;
  insert_to_column (0, place);
}

forward_one_sexp()
{
  int start = point;
  int level = 0;

  /* Jump over whitespace and close parentheses.
   Abort if we reach the end of the buffer.
   Leave point on the first non-whitespace-or-close character we see. */

  if (!re_search(1, NOT_WHITESPACE_OR_CLOSE)) {
    say("At end of buffer");
    point = start;
    return 1;
  }
  point = matchstart;

  /* Loop skipping forward over Lisp tokens.  The variable LEVEL keeps
   track of the current nesting level. */

  while (1) {
    switch (curchar()) {

      case OPEN_PAREN:
        point++;
        level++;
        break;

      case CLOSE_PAREN:
        point++;
        if (level > 0) level--;
        break;

      case ';':
        nl_forward();
        goto next_token;

      case '\'':
      case '`':
      case ',':
        point++;
        goto next_token;

      case '#':
        point++;
        switch (curchar()) {

          case OPEN_PAREN:                    /* Scheme and CL #( */
            point++;
            level++;
            break;

          case '\\':                          /* Scheme and CL #\ */
            point++;
            if (curchar() == '\\')            /* Must treat #\\ specially */
              point++;
            else if (re_search(1, LISP_BREAK))
              point--;
            break;

          case 'T':                           /* Scheme #T and #F */
          case 't':
          case 'F':
          case 'f':
            point++;
            break;

          case 'B':                           /* Scheme and CL #B #O #X */
          case 'b':
          case 'O':
          case 'o':
          case 'X':
          case 'x':

          case 'D':                           /* Scheme #D */
          case 'd':

          case 'R':
          case 'r':                           /* CL #R */

          case '\'':                          /* CL #' #. #, #+ #- */
          case '.':
          case ',':
          case '+':
          case '-':
            point++;
            goto next_token;

          case '|':                           /* CL #| ... |# */
            if (!re_search(1, "%|#")) {
              say("Unmatched comment");
              point = start;
              return 0;
            }
            break;

          default:
            break;
        }
        break;

        /* We're inside a string.  Search for the next two-character sequence
         where the first character is not backslash and the second character
         is double quote. */

      case '"':
        re_search(1, "[^\\]\"");
        break;

        /* We're inside a quoted symbol.  Search for the next two-character
         sequence where the first character is not backslash and the second
         character is vertical bar. */

      case '|':
        re_search(1, "[^\\]%|");
        break;

        /* We're inside an ordinary symbol.  Search for the next two-character
         sequence where the first character is not a backslash and the second
         character is a Lisp token break character. */

      default:
        if (re_search(1, LISP_BREAK))
          point--;
        break;
    }

    if (level == 0) return 1;

    /* Skip over whitespace to find the start of the next token.  Leave point
     on the break character, if we find one, or at the end of the buffer. */

  next_token:
    if (!re_search(1, NOT_WHITESPACE))
      break;
    point = matchstart;
  }

  say("Unmatched parentheses");
  point = start;
  return 0;
}

backward_one_sexp()
{
  int start = point;
  int level = 0;

  /* Jump over whitespace and open parentheses.
   Abort if we reach the beginning of the buffer. */

  if (!re_search(-1, NOT_WHITESPACE_OR_OPEN)) {
    say("At beginning of buffer");
    point = start;
    return 1;
  }

  /* Loop skipping backward over Lisp tokens.  The variable LEVEL keeps
   track of the current nesting level. */

  while (1) {
    switch (curchar()) {

      /* If we see a slashified open paren at the end of a token,
       it is either a character constant or a symbol ending in a
       slashified paren. */

      case CLOSE_PAREN:
        if (character(point-1) == '\\') {
          re_search(-1, LISP_BREAK);
          point++;
        }
        else
          level++;
        break;

      case OPEN_PAREN:
        if (character(point-1) == '\\') {
          re_search(-1, LISP_BREAK);
          point++;
        }
        else {
          if (level > 0) level--;
          if (index("'`,#", character(point-1)))
            point--;
          else if (index("+-", character(point-1)) &&
                   (character(point-2) == '#'))
            point -= 2;
        }
        break;

        /* We're inside a string.  Search for the next two-character sequence
         where the first character is not backslash and the second character
         is double quote. */

      case '"':
        re_search(-1, "[^\\]\"");
        point++;
        break;

        /* We're inside a quoted symbol.  Search for the next two-character
         sequence where the first character is not backslash and the second
         character is vertical bar. */

      case '|':
        re_search(-1, "[^\\]%|");
        if (!index("'`,", curchar()))
          point++;
        break;

      default:

        /* We're inside a comment. */

        if ((curchar() == '#') && (character(point-1) == '|')) {
          if (!re_search(-1, "#%|")) {
            say("Unmatched comment");
            point = start;
            return 0;
          }
        }

        /* We're inside an ordinary symbol.  Search for the next two-character
         sequence where the first character is not a backslash and the second
         character is a Lisp token break character. */

        else if (re_search(-1, LISP_BREAK))
          point = matchstart;

        /* We must treat the case of a single Lisp break character at the
         beginnning of the buffer specially, since we won't find it as
         a two-character sequence. */

        else if (index("()|\"#", curchar()))
          point++;
        break;
    }

    if (level == 0) return 1;

    if (!re_search(-1, NOT_WHITESPACE))
      break;
  }

  say("Unmatched parentheses");
  point = start;
  return 0;
}

command forward_sexp() on lisp_tab[ALT(CTRL('F'))]
{
  if (iter < 0)
    while (iter++ < 0)
      backward_one_sexp();
  else
    while (iter-- > 0)
      forward_one_sexp();
}

command backward_sexp() on lisp_tab[ALT(CTRL('B'))]
{
  if (iter < 0)
    while (iter++ < 0)
      forward_one_sexp();
  else
    while (iter-- > 0)
      backward_one_sexp();
}

command kill_sexp() on lisp_tab[ALT(CTRL('K'))]
{
  int start = point;
  forward_sexp();
  do_save_kill(start, point);
}

command up_sexp() on lisp_tab[ALT(CTRL('U'))]
{
  while (1) {
    if (!re_search(-1, "[^ \t\n]"))
      break;
    if (curchar() == OPEN_PAREN)
      break;
    point++;
    if (!backward_one_sexp())
      break;
    if (current_column() == 0)
      break;
  }
}

/* What should this command do when the cursor is on an open paren?
 At first I thought it should move forward one character and then
 try to go down a level.  However, this means that C-M-D followed
 by C-M-U doesn't leave you where you began.  So, now I have it
 defined to just go forward one character. */

command down_sexp() on lisp_tab[ALT(CTRL('D'))]
{
  int start = point;
  /* Must treat this as a special case since the re_search
   will only look for two character sequences. */
  if (curchar() == OPEN_PAREN) {
    point++;
    return;
  }
  if (!re_search(1, "[^\\][()]"))
    return;
  if (character(point-1) == CLOSE_PAREN) {
    point = start;
  }
}

command begin_defun() on lisp_tab[ALT(CTRL('A'))]
{
  while (1) {
    if (!search(-1, "("))
      break;
    if (current_column() == 0)
      break;
  }
}

lisp_compute_indent()
{
  int start, indent;
  start = point;

  /* Find first open or close paren above the current line. */
  to_begin_line();
  if (!re_search(-1, "[()]")) {
    /* No parens at all before point.  
     Leave point at the end of the indentation on the current line. */
    point = start;
    to_indentation();
    indent = current_column();
    point = start;
    return indent;
  }

  /* Skip backward over the preceding S-expression.  Then search backward
   for either the beginning of the line or the operator of the current form.
   To correctly indent DO, DO*, UNWIND-PROTECT, and LOOP, it is necessary to
   always search for the operator of the current form. */

  if (curchar() == CLOSE_PAREN)
    while (1) {
      point++;
      backward_one_sexp();
      indent = current_column();
      if (!re_search(-1, "[()\n]")) {
        point = start;
        return indent;
      }
      if (curchar() == '\n') {
        point++;
        to_indentation();
        indent = current_column();
        point = start;
        return indent;
      }
      if (curchar() == OPEN_PAREN)
        break;
    }

    /* The point is just before an open paren.  Find the indentation of the
     first S-expression following the operator.  Also check the operator name
     for certain special forms. */

    {
      int operator_start, operator_end;
      int open_paren_column = current_column();
      point++;
      operator_start = point;
      re_search(1, "[ \t\n(]");
      point--;
      operator_end = point;
      if ((curchar() == ' ') || (curchar() == '\t')) {
        re_search(1, "[^ \t]");
        point--;
      }
      if (curchar() == '\n')
        indent = open_paren_column + lisp_indent;
      else {
        /* It might be better to do this with a regexp. */
        char operator[80];
        grab(operator_start, operator_end, operator);
        if (!(   strnfcmp(operator, "def", 3)
              && strnfcmp(operator, "let", 3)
              && strnfcmp(operator, "with", 4)
              && strfcmp(operator, "case")
              && strfcmp(operator, "flet")
              && strfcmp(operator, "when")
              && strfcmp(operator, "ccase")
              && strfcmp(operator, "ecase")
              && strfcmp(operator, "labels")
              && strfcmp(operator, "lambda")
              && strfcmp(operator, "unless")
              && strfcmp(operator, "dolist")
              && strfcmp(operator, "dotimes")
              && strfcmp(operator, "macrolet")
              && strfcmp(operator, "ctypecase")
			  && strfcmp(operator, "etypecase")
			  && strfcmp(operator, "de")
              ))
          indent = open_paren_column + lisp_indent;
        else
          indent = current_column();
      }
    }

    point = start;
    return indent;
}

handle_lisp_indent()
{
  int pos = size() - point;
  int indent = lisp_compute_indent ();

  to_indentation ();
  if (current_column () != indent)
    my_to_column (indent);
  if (size() - pos > point)
    point = size() - pos;
}

lisp_indent_cmd() on lisp_tab['\t']
{
  standard_tab_cmd(handle_lisp_indent, lisp_indent, 1);
}

command indent_sexp() on lisp_tab[ALT(CTRL('Q'))]
{
  spot start = alloc_spot();
  spot end = alloc_spot();
  forward_one_sexp();
  *end = point;
  point = *start;
  while (1) {
    if (!nl_forward())
      break;
    if (point >= *end)
      break;
    my_to_column(lisp_compute_indent());
  }
  point = *start;
  free_spot (start);
  free_spot (end);  
}

/* The command show_matching_delimiter tests to see if move_level
 returns 1, and only then does a show_line.  Should we do the same? */

command show_matching_paren() on lisp_tab[CLOSE_PAREN]
{
  int start;
  normal_character();
  start = point;
  say("");
  backward_one_sexp();
  show_line();
  point = start;
}

command insert_parens() on lisp_tab[ALT(OPEN_PAREN)]
{
  stuff("()");
  point--;
}

command move_over_close_paren() on lisp_tab[ALT(CLOSE_PAREN)]
{
  re_search(1, ")");
}

do_lisp_indent()
{
  my_to_column(lisp_compute_indent());
}

lisp_color_c_range(int from, int to)
{
  save_var point, narrow_start = from, narrow_end = size() - to;
  color_c_range(from, to);
}


color_lisp_range(from, to)
{
  int t = -1;

  if (from >= to)
	return to;
  save_var point, matchstart, matchend, case_fold = 1;
  point = to;				// Color entire lines.
  nl_forward();
  to = point;
  save_var narrow_end = size() - to;
  point = from;
  to_begin_line();
  while (parse_string(-1, "\\\n") && matchend < point) {
	point = matchend;
	to_begin_line();
  }
  from = point;
  if (!re_search(-1, "#{"))		// If embedded C, go before it.
	point = from;			// None, so don't need to go back.
  set_character_color(from, to, -1);

  while (point < to) {	// Search for special char or string or comment.
	if (!re_search(1, "[;#\"()]|/%*")) {
	  t = size();
	  break;
	}
	t = matchstart;
	if (character(t) == ';') {
	  nl_forward();
	  set_character_color(t, point, color_class lisp_comment);
	} else if (character(t) == '/') {
	  search(1, "*/");
	  set_character_color(t, point, color_class lisp_comment);
	} else if (character(t) == '"') {
	  re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
	  set_character_color(t, point, color_class lisp_string);
	} else if (character(t) == '#') {
	  switch (character(t + 1)) {
		case '{':;		// C code
		  search(1, "#}");
		  set_character_color(t, point, color_class lisp_special);
		  lisp_color_c_range(t + 2, point - 2);
		  break;
		case '?':
		  set_character_color(t, give_end_line(), color_class lisp_special);
		  break;
		default:
		  set_character_color(t, point, -1);
		  break;
	  }
	} else if (character(t) == '(') {
	  set_character_color(t, point, color_class lisp_punctuation);
	  re_search(1, "[^ \n\"()#]*");
	  set_character_color(matchstart, point, color_class lisp_function);
	} else if (character(t) == ')') {
	  set_character_color(t, point, color_class lisp_punctuation);
	} else
	  set_character_color(t, point, -1);
  }
  if (to < t)
	set_character_color(to, t, -1);
  return point;
}

command lisp_mode ()
{
  mode_default_settings();
  major_mode = strkeep("Lisp");
  mode_keys = lisp_tab;
  strcpy(comment_start, ";[ \t]*");
  strcpy(comment_pattern, ";.*$");
  strcpy(comment_begin, "; ");
  strcpy(comment_end, "");
  comment_column = 0;
  recolor_range = color_lisp_range;	// set up coloring rules
  recolor_from_here = recolor_by_lines;
  if (want_code_coloring)		// maybe turn on coloring
	when_setting_want_code_coloring();
  indenter = do_lisp_indent;
  auto_indent = 1;
  margin_right = 79;
  word_pattern = lisp_word_pattern;
  buffer_maybe_break_line = generic_maybe_break_line;
  fill_mode = (misc_language_fill_mode & 128) != 0;
  if (auto_show_lisp_delimiters)
	  auto_show_matching_characters = lisp_auto_show_delim_chars;
  try_calling("lisp-mode-hook");
  make_mode();
}

when_loading()
{
  lisp_tab[ALT('q')] = (short) fill_comment;
  lisp_tab[ALT('Q')] = (short) prefix_fill_paragraph;
}

suffix_lsp()  { lisp_mode(); }
suffix_scm()  { lisp_mode(); }
suffix_lsh()  { lisp_mode(); }

tag_mode_lisp()
{
  int old_case_fold = case_fold;
  int end;
  int start;
  char func[201];

  case_fold = 1;
  while (re_search (1, lisp_tag_pattern))
  {                                   
    point--;
    forward_one_sexp ();
    forward_one_sexp ();
    end = point;
    backward_one_sexp ();
    start = point;
    if (character (start) == '\'') start++;
    else
    {
      if (character(start) == OPEN_PAREN)
      {
        point++;
        forward_one_sexp ();
        end = point;
        backward_one_sexp ();
        start = point;
      }
    }
    grab (start, end, func);
    add_tag (func, start);
    point = end;
  }
  case_fold = old_case_fold;
}

// Local Variables:
// c-indent: 2
// End: 
